"""UCI Hour Record handler for trackmeet."""

import gtk
import glib
import gobject
import logging
import os

import metarace
from metarace import timy
from metarace import scbwin
from metarace import tod
from metarace import unt4
from metarace import uiutil
from metarace import jsonconfig
from metarace import strops
from metarace import report

_log = logging.getLogger(u'metarace.hour')
_log.setLevel(logging.DEBUG)

# config version string
EVENT_ID = 'hourrec'

# run clock up to this long after expiry of hour
MAX_AFTER = tod.tod(u'10:00')

# scb function key mappings (also trig announce)
key_timerwin = 'F6'  # re-show timing window
key_results = u'F4'

# timing function key mappings
key_armstart = 'F5'  # arm start -> countdown
key_armfinish = 'F9'  # override finish trigger

# extended function key mappings
key_reset = 'F5'  # + ctrl for clear/abort
key_falsestart = 'F6'  # + ctrl for false start
key_abort = 'F7'  # + ctrl abort A


class hourrec(object):
    """Data handling for Hour record."""

    def key_event(self, widget, event):
        """Race window key press handler."""
        if event.type == gtk.gdk.KEY_PRESS:
            key = gtk.gdk.keyval_name(event.keyval) or 'None'
            if event.state & gtk.gdk.CONTROL_MASK:
                if key == key_reset:  # override ctrl+f5
                    self.toidle()
                    return True
                elif key == key_results:
                    glib.idle_add(self.delayed_announce)
                    return True
                elif key == key_falsestart:
                    self.falsestart()
                    return True
                elif key == key_abort:
                    self.abortrider()
                    return True
            elif key[0] == 'F':
                if key == key_armstart:
                    self.armstart()
                    return True
                elif key == key_armfinish:
                    self.armfinish()
                    return True
                elif key == key_timerwin or key == key_results:
                    self.showtimerwin()
                    glib.idle_add(self.delayed_announce)
                    return True
        return False

    def recalc(self):
        """Recalc runtime from current state."""
        self.statusline.set_text(u'')
        self.curspeed = None
        elap = None
        if self.finish is None and self.elapsed == u'' and self.start is not None and len(
                self.splitlist) > 0:
            # elapsed time at last completed lap
            self.elapsed = (self.splitlist[-1] - self.start).rawtime(0)
        if self.lstart is not None:
            elap = tod.now() - self.lstart  # use ltime for roll-over
        if self.finish is not None:  # rider has crossed line after time
            self.TC = self.lapcount - 1
            self.TTC = None
            if len(self.splitlist) > 2:
                belltime = self.splitlist[-2]
                laststart = self.splitlist[-3]
                # time of the last complete lap
                self.TTC = belltime - laststart
                # time remaining to ride at the beginning of the last lap
                self.TRC = self.reclen - (belltime - self.start)
                # additional distance
                self.DiC = self.LPi * float(self.TRC.timeval) / float(
                    self.TTC.timeval)
                # distance covered in the hour
                self.D = int(self.DiC + self.TC * self.LPi)
                self.compute = u'D={0}\u2006m, LPi\u00d7TC={1:0.1f}\u2006m, LPi={2:0.1f}\u2006m, TC={3}, DiC={4:0.1f}\u2006m, TTC={5}, TRC={6}'.format(
                    self.D, self.LPi * self.TC, self.LPi, self.TC, self.DiC,
                    self.TTC.rawtime(3), self.TRC.rawtime(3))
                self.infoline.set_text(
                    u'D: {0:0.3f}\u2006km, DiC: {1:0.1f}\u2006m'.format(
                        self.D / 1000.0, self.DiC))
                _log.debug(self.compute)
            self.projection = None
            self.statusline.set_text(u'{} Laps'.format(self.TC))
        else:
            self.TC = self.lapcount
            self.statusline.set_text(u'{} Laps'.format(self.TC))
            if elap > self.reclen:  # time has expired
                self.TTC = None
                if len(self.splitlist) > 1:
                    belltime = self.splitlist[-1]
                    laststart = self.splitlist[-2]
                    # time of the last complete lap
                    self.TTC = belltime - laststart
                    # time remaining to ride at the beginning of the last lap
                    self.TRC = self.reclen - (belltime - self.start)
                    # additional distance
                    self.DiC = self.LPi * float(self.TRC.timeval) / float(
                        self.TTC.timeval)
                    # distance covered in the hour
                    self.D = int(self.DiC + self.TC * self.LPi)
                    self.compute = u'D={0}\u2006m, LPi\u00d7TC={1:0.1f}\u2006m, LPi={2:0.1f}\u2006m, TC={3}, DiC={4:0.1f}\u2006m, TTC={5}, TRC={6}'.format(
                        self.D, self.LPi * self.TC, self.LPi, self.TC,
                        self.DiC, self.TTC.rawtime(3), self.TRC.rawtime(3))
                    self.infoline.set_text(
                        u'D: {0:0.3f}\u2006km, DiC: {1:0.1f}\u2006m'.format(
                            self.D / 1000.0, self.DiC))
                    _log.debug(self.compute)
                self.projection = None
            else:
                # update projection (use 3 lap average speed)
                if len(self.splitlist) > 3:
                    et = self.splitlist[-1] - self.start
                    rt = self.reclen - et
                    di = int(self.LPi * self.lapcount)
                    pi = self.splitlist[-4]
                    pd = self.splitlist[-1]
                    pr = (self.LPi * 3) / (float(pd.timeval) -
                                           float(pi.timeval))
                    self.curspeed = 3.6 * pr
                    de = di + int(float(rt.timeval) * pr)
                    _log.debug(u'Projected final dist: %r', de)
                    self.infoline.set_text(
                        u'Speed: {0:0.1f}\u2006km/h, Est: {1:0.2f}\u2006km'.
                        format(self.curspeed, de / 1000.0))
                    if de < self.maxproj and de > self.minproj:
                        self.projection = de
                    else:
                        self.projection = None
        # update telegraph outputs
        if self.D is not None and self.D > 0:
            self.meet.cmd_announce(u'distance', str(self.D))
        else:
            self.meet.cmd_announce(u'distance', None)
        if self.projection is not None:
            self.meet.cmd_announce(u'projection', unicode(self.projection))
        else:
            self.meet.cmd_announce(u'projection', None)
        if self.curspeed is not None:
            self.meet.cmd_announce(u'curspeed',
                                   '{0:0.1f}'.format(self.curspeed))
        else:
            self.meet.cmd_announce(u'curspeed', None)

    def update_expander_lbl_cb(self):
        """Update race info expander label."""
        self.info_expand.set_label(self.meet.infoline(self.event))

    def loadconfig(self):
        """Load race config from disk."""
        self.model.clear()
        cr = jsonconfig.config({
            u'event': {
                u'id': EVENT_ID,
                u'comments': [],
                u'riderstr': u'',
                u'wallstart': None,
                u'start': None,
                u'showinfo': True,
                u'finish': None,
                u'lstart': None,
                u'target': None,
                u'reclen': '1h00:00',
                u'minlap': '14.0',
                u'record': None,
                u'projlap': 12,
                u'lpi': None,
                u'minproj': 30000,
                u'maxproj': 60000,
                u'lapcount': 0,
                u'splitlist': []
            }
        })
        cr.add_section(u'event')
        if os.path.exists(self.configfile):
            try:
                with open(self.configfile, 'rb') as f:
                    cr.read(f)
            except Exception as e:
                _log.error(u'Unable to read config: %s', e)
        else:
            _log.info(u'%r not found, loading defaults', self.configfile)

        # race infos
        self.comments = cr.get(u'event', u'comments')
        self.update_expander_lbl_cb()
        self.info_expand.set_expanded(
            strops.confopt_bool(cr.get(u'event', u'showinfo')))

        self.riderstr = cr.get(u'event', u'riderstr')

        self.reclen = tod.mktod(cr.get(u'event', u'reclen'))
        if self.reclen is None:
            _log.error(u'Invalid record length: Reset to 1h00:00')
            self.reclen = tod.tod(u'1h00:00')
        self.minlap = tod.mktod(cr.get(u'event', u'minlap'))
        if self.minlap is None:
            _log.error(u'Invalid min lap time: Reset to 14.0s')
            self.minlap = tod.tod(u'14.0')
        self.start = tod.mktod(cr.get(u'event', u'start'))
        self.finish = tod.mktod(cr.get(u'event', u'finish'))
        self.lstart = tod.mktod(cr.get(u'event', u'lstart'))
        self.wallstart = tod.mktod(cr.get(u'event', u'wallstart'))

        self.target = strops.confopt_posint(cr.get(u'event', u'target'))
        self.record = strops.confopt_posint(cr.get(u'event', u'record'))
        chklap = float(self.meet.tracklen_n) / float(self.meet.tracklen_d)
        self.LPi = strops.confopt_float(cr.get(u'event', u'lpi'), chklap)
        self.projlap = strops.confopt_posint(cr.get(u'event', u'projlap'), 12)
        self.minproj = strops.confopt_posint(cr.get(u'event', u'minproj'),
                                             30000)
        self.maxproj = strops.confopt_posint(cr.get(u'event', u'maxproj'),
                                             60000)
        self.lapcount = strops.confopt_posint(cr.get(u'event', u'lapcount'))
        self.splitlist = []
        lc = 0
        pt = self.start
        for lt in cr.get(u'event', u'splitlist'):
            nlt = tod.mktod(lt)
            if nlt is not None:
                lc += 1
                self.splitlist.append(nlt)
                if self.start is not None:
                    self.addlapline(nlt, pt, lc, False)
                pt = nlt
        # check count of laps agains splits
        if len(self.splitlist) != self.lapcount:
            _log.error(u'SPLIT LIST != LAPCOUNT')

        # arm the front straight
        if self.winopen:
            self.meet.main_timer.arm(2)
            self.meet.main_timer.armlock(True)
            self.meet.scbwin = scbwin.scbtt(scb=self.meet.scb,
                                            header=self.event[u'pref'],
                                            subheader=self.event[u'info'])
            self.meet.scbwin.reset()
        # recalc
        self.recalc()

        # then set UI
        if self.start is None:
            self.toidle()
        else:
            if self.D is not None and self.D > 0:
                self.tofinish()
            else:
                self.torunning()

    def addlapline(self, split, lastsplit, count, scroll=True):
        """Insert a lap line on the ui view."""
        ctxt = unicode(count)
        selap = split - self.start
        dtxt = u'[final lap]'
        if selap < self.reclen:
            dtxt = u'{0:0.3f}\u2006km'.format(count * self.LPi / 1000.0)
        stxt = (split - self.start).rawtime(3)
        ltxt = (split - lastsplit).rawtime(3)
        self.model.insert(0, [ctxt, dtxt, stxt, ltxt])
        if scroll:
            self.view.scroll_to_cell(0)

    def saveconfig(self):
        """Save race to disk."""
        if self.readonly:
            _log.error('Attempt to save readonly ob.')
            return
        cw = jsonconfig.config()
        cw.add_section(u'event')

        cw.set(u'event', u'showinfo', self.info_expand.get_expanded())
        cw.set(u'event', u'riderstr', self.riderstr)
        cw.set(u'event', u'reclen', self.reclen.rawtime())
        cw.set(u'event', u'minlap', self.minlap.rawtime())
        if self.start is not None:
            cw.set(u'event', u'start', self.start.rawtime())
        if self.lstart is not None:
            cw.set(u'event', u'lstart', self.lstart.rawtime())
        if self.wallstart is not None:
            cw.set(u'event', u'wallstart', self.wallstart.rawtime())
        if self.finish is not None:
            cw.set(u'event', u'finish', self.finish.rawtime())

        cw.set(u'event', u'lpi', self.LPi)
        cw.set(u'event', u'target', self.target)
        cw.set(u'event', u'record', self.record)
        cw.set(u'event', u'projlap', self.projlap)
        cw.set(u'event', u'minproj', self.minproj)
        cw.set(u'event', u'maxproj', self.maxproj)
        cw.set(u'event', u'lapcount', self.lapcount)
        cw.set(u'event', u'comments', self.comments)

        slout = []
        for lt in self.splitlist:
            slout.append(lt.rawtime())
        cw.set(u'event', u'splitlist', slout)

        cw.set(u'event', u'id', EVENT_ID)
        _log.debug(u'Saving event config %r', self.configfile)
        with metarace.savefile(self.configfile) as f:
            cw.write(f)

    def startlist_report(self, program=False):
        """Return a startlist report."""
        ret = []
        cnt = 0
        sec = report.bullet_text()
        sec.heading = u' '.join([self.event[u'pref'],
                                 self.event[u'info']]).strip()
        substr = u' '.join([self.event[u'dist'], self.event[u'prog']]).strip()
        if substr:
            sec.subheading = substr
        if self.event[u'reco']:
            sec.footer = self.event[u'reco']
        # the rider free-form string
        if self.riderstr:
            sec.lines.append([u'', self.riderstr])
        if self.record is not None:
            tstr = u'{0:0.3f}\u2006km'.format(self.record / 1000.0)
            sec.lines.append([u'', u'Record: ' + tstr])
        if self.target is not None:
            tstr = u'{0:0.3f}\u2006km'.format(self.target / 1000.0)
            sec.lines.append([u'', u'Target: ' + tstr])
        if self.wallstart is not None:
            sec.lines.append(
                [u'', u'Start Time: ' + self.wallstart.meridiem(secs=False)])
        ret.append(sec)
        return ret

    def get_startlist(self):
        """Return a list of bibs in the rider model."""
        ret = []
        for r in self.riders:
            ret.append(r[COL_BIB])
        return ' '.join(ret)

    def delayed_announce(self):
        """Initialise the announcer's screen after a delay."""
        if self.winopen:
            _log.debug(u'TODO delayed_announce')

    def shutdown(self, win=None, msg=u'Exiting'):
        """Terminate race object."""
        rstr = u''
        if self.readonly:
            rstr = u'readonly '
        _log.debug(u'Shutdown %sevent %s: %s', rstr, self.evno, msg)
        if not self.readonly:
            self.saveconfig()
        self.winopen = False

    def do_properties(self):
        """Run race properties dialog."""
        prfile = os.path.join(metarace.UI_PATH, u'hour_properties.ui')
        _log.debug(u'Building event properties from %r', prfile)
        b = gtk.Builder()
        b.add_from_file(prfile)
        dlg = b.get_object(u'properties')
        dlg.set_transient_for(self.meet.window)
        rl = b.get_object(u'recordlen_entry')
        rl.set_text(self.reclen.rawtime(0))
        ri = b.get_object(u'ridername_entry')
        ri.set_text(self.riderstr)
        st = b.get_object(u'start_entry')
        if self.wallstart is not None:
            st.set_text(self.wallstart.rawtime(0))
        rc = b.get_object(u'record_entry')
        if self.record is not None:
            rc.set_text(unicode(self.record))
        tg = b.get_object(u'target_entry')
        if self.target is not None:
            tg.set_text(unicode(self.target))
        response = dlg.run()
        if response == 1:  # id 1 set in glade for "Apply"
            _log.debug(u'Updating event properties')
            # update setting
            self.riderstr = ri.get_text().decode(u'utf-8')
            checklen = tod.mktod(rl.get_text().decode(u'utf-8'))
            if checklen is not None:
                self.reclen = checklen
            else:
                _log.error(u'Invalid record len %r, using default: 1h00:00',
                           rl.get_text().decode(u'utf-8'))
                self.reclen = tod.mktod(u'1h00:00')
            self.wallstart = tod.mktod(st.get_text().decode(u'utf-8'))
            self.target = strops.confopt_posint(tg.get_text().decode(u'utf-8'),
                                                None)
            self.record = strops.confopt_posint(rc.get_text().decode(u'utf-8'),
                                                None)
            glib.idle_add(self.delayed_announce)
        else:
            _log.debug(u'Edit event properties cancelled')

        # if prefix is empty, grab input focus
        if not self.prefix_ent.get_text():
            self.prefix_ent.grab_focus()
        dlg.destroy()

    def result_gen(self):
        """Generator function to export a final result."""
        yield [u'', u'', u'', u'']

    def result_report(self, recurse=False):
        """Return a list of report sections containing the race result."""
        ret = []
        sec = report.bullet_text()
        sec.heading = u' '.join([self.event[u'pref'],
                                 self.event[u'info']]).strip()
        substr = u' '.join([self.event[u'dist'], self.event[u'prog']]).strip()
        if substr:
            sec.subheading = substr
        if self.event[u'reco']:
            sec.footer = self.event[u'reco']

        sec.lines.append([u'', self.riderstr])
        # Distance measure - result if D available
        #if self.finish is not None:
        if self.D is not None and self.D > 0:
            dstr = u'{0:0.3f}\u2006km'.format(self.D / 1000.0)
            sec.lines.append([u'', u'Final distance: ' + dstr])
            if self.record is not None:
                if self.D > self.record:
                    sec.lines.append([
                        '', u'New record by {} metre{}'.format(
                            self.D - self.record,
                            strops.plural(self.D - self.record))
                    ])
                else:
                    tstr = u'{0:0.3f}\u2006km'.format(self.record / 1000.0)
                    sec.lines.append([
                        u'', u'{} metre{} short of existing record: '.format(
                            self.record - self.D,
                            strops.plural(self.record - self.D)) + tstr
                    ])
            sec.lines.append([u'', u'Complete laps: ' + unicode(self.TC)])
            if self.compute:
                sec.lines.append(
                    [u'', u'Additional distance: {0:0.1f} m'.format(self.DiC)])
                sec.lines.append([u'', self.compute])
        else:
            if self.record is not None:
                tstr = u'{0:0.3f}\u2006km'.format(self.record / 1000.0)
                sec.lines.append([u'', u'Record: ' + tstr])
            if self.target is not None:
                tstr = u'{0:0.3f}\u2006km'.format(self.target / 1000.0)
                sec.lines.append([u'', u'Target: ' + tstr])
            if self.wallstart is not None:
                sec.lines.append([
                    u'', u'Start Time: ' + self.wallstart.meridiem(secs=False)
                ])
            if self.projection is not None:
                tstr = u'{0:0.3f}\u2006km'.format(self.projection / 1000.0)
                sec.lines.append([u'', u'Projection: ' + tstr])
            if self.lapcount > 0:
                sec.lines.append([u'', u'Elapsed: ' + self.elapsed])
                sec.lines.append([u'', u'Laps: ' + unicode(self.lapcount)])
        ret.append(sec)
        if self.start is not None and self.lapcount > 0:
            sec = report.threecol_section()
            sec.subheading = u'Lap Times'
            sec.lines = []
            lt = self.start
            count = 1
            ld = 0
            for st in self.splitlist:
                laptime = st - lt
                split = st - self.start
                lstr = u'{}'.format(count)
                nd = int(0.010 + self.LPi * count / 1000.0)
                if nd != ld:
                    lstr += u'  / {}\u2006km'.format(nd)
                    ld = nd
                sec.lines.append([
                    u'', u'', lstr, u'',
                    laptime.rawtime(3),
                    split.rawtime(3)
                ])
                lt = st
                count += 1
            ret.append(sec)
        if len(self.comments) > 0:
            sec = report.bullet_text()
            sec.subheading = u'Decisions of the commisaires panel'
            for c in self.comments:
                sec.lines.append([None, c])
            ret.append(sec)
        return ret

    def addrider(self, bib='', info=None):
        return None

    def editent_cb(self, entry, col):
        """Shared event entry update callback."""
        if col == u'pref':
            self.event[u'pref'] = entry.get_text().decode(u'utf-8')
        elif col == u'info':
            self.event[u'info'] = entry.get_text().decode(u'utf-8')
        self.update_expander_lbl_cb()

    def split_trig(self, t):
        """Register lap trigger."""
        if self.start is not None:
            lastlap = None
            lasttwo = None
            if len(self.splitlist) > 0:
                lastlap = self.splitlist[-1]
            else:
                lastlap = self.start
            if len(self.splitlist) > 2:
                lasttwo = t - self.splitlist[-2]
            laptime = t - lastlap
            if laptime > self.minlap:
                self.laptimedirty = True
                elap = t - self.start
                if elap <= self.reclen:
                    self.lapcount += 1
                    self.splitlist.append(t)
                    self.addlapline(t, lastlap, self.lapcount)
                    self.recalc()
                    self.curlap = laptime
                    glib.idle_add(self.scblap)
                    _log.info(u'Lap %r: %s @ %s', self.lapcount,
                              laptime.rawtime(2), t.rawtime(2))
                    if laptime < 60:
                        self.lastlapstr = laptime.rawtime(1)
                    else:
                        self.lastlapstr = laptime.rawtime(0)
                    if (elap + laptime) > self.reclen:
                        self.tofinal()
                    elif lasttwo is not None:
                        if (elap + lasttwo) > self.reclen:
                            self.tobell()
                else:
                    if self.finish is None:
                        self.lapcount += 1
                        self.splitlist.append(t)
                        self.addlapline(t, lastlap, self.lapcount)
                        self.finish = t
                        self.curlap = laptime
                        self.recalc()
                        self.tofinish()
                        _log.info(u'Final Lap Completed.')
                    else:
                        _log.info(u'Duplicate finish pass.')
                # and ask meet for an export
                self.meet.delayed_export()
            else:
                _log.info(u'Ignored short lap.')
        else:
            _log.info(u'Ignored trig without start.')

    def scblap(self):
        # output to main scoreboard
        if type(self.meet.scbwin) is scbwin.scbtt:
            self.meet.scbwin.setline1(
                strops.truncpad(
                    u'Elapsed: ', self.meet.scb.linelen - 12, align='r') +
                strops.truncpad(self.elapsed, 12))

            if self.lapcount > 0:
                ## display kilometre updates elsewhere?
                if self.lapcount > 3:
                    pass
                    #kilo = self.lapcount // 4
                    #self.meet.scbwin.setr1(u'Lap {0}, {1}\u2006km:'.format(
                    #self.lapcount,kilo))
                #else:
                self.meet.scbwin.setr1(u'Lap {0}:'.format(self.lapcount))
                self.meet.scbwin.sett1(self.lastlapstr)
            else:
                self.meet.scbwin.setline2(u'')
                self.meet.scbwin.setr1(u'')
                self.meet.scbwin.sett1(u'')
            if self.record is not None:
                self.meet.scbwin.setline2(
                    strops.truncpad(
                        u'Record: ', self.meet.scb.linelen - 12, align='r') +
                    strops.truncpad(
                        u'{0:0.3f}\u2006km'.format(self.record / 1000.0), 12))
            elif self.target is not None:
                self.meet.scbwin.setline2(
                    strops.truncpad(
                        u'Target: ', self.meet.scb.linelen - 12, align='r') +
                    strops.truncpad(
                        u'{0:0.3f}\u2006km'.format(self.target / 1000.0), 12))
            if self.lapcount > self.projlap and self.projection is not None:
                self.meet.scbwin.setr2(u'Projection:')
                self.meet.scbwin.sett2(u'{0:0.2f}\u2006km'.format(
                    self.projection / 1000.0))
            else:
                self.meet.scbwin.setr2(u'')
                self.meet.scbwin.sett2(u'')
            self.meet.scbwin.update()

        # telegraph outputs
        self.meet.cmd_announce(u'lapcount', unicode(self.lapcount))
        if self.laptimedirty:
            self.meet.cmd_announce(u'laptime', self.lastlapstr)
            self.laptimedirty = False
        self.meet.cmd_announce(u'elapsed', self.elapsed)

        # on the gemini - use B/T dual timer mode
        self.meet.gemini.set_bib(str(self.lapcount), 0)
        self.meet.gemini.set_time(self.lastlapstr, 0)
        self.meet.gemini.set_time(self.elapsed, 1)
        self.meet.gemini.show_dual()

        return False

    def timercb(self, e):
        """Handle a timer event."""
        chan = timy.chan2id(e.chan)
        if chan == 0:
            if self.start is None:
                self.lstart = tod.now()
                self.start = e
                self.torunning()
                _log.info(u'Set Start: %s', e.rawtime())
            else:
                _log.info(u'Spurious start trig: %s', e.rawtime())
        elif chan == 2:
            self.split_trig(e)
        else:
            _log.info(u'Trigger: %r @ %s', chan, e.rawtime())
        return False

    def timeout(self):
        """Update scoreboard and respond to timing events."""
        if not self.winopen:
            return False
        now = tod.now()
        nelap = self.elapsed
        dofinishtxt = False
        dostarttxt = False
        setarmfinish = False
        # determine the new elapsed time in secs
        if self.lstart is not None and self.finish is None:
            tot = now - self.lstart
            #self.elaptod = tot
            if tot >= (self.reclen + MAX_AFTER):
                nelap = u'--:--'
            else:
                nelap = tot.rawtime(0)
            if tot > self.reclen:  # time has expired
                setarmfinish = True

        elif self.finish is not None:
            # the hour is over
            #nelap = u'60:00'
            nelap = u'--:--'
            dofinishtxt = True
        else:
            # Before Start
            dostarttxt = True
            if self.wallstart is not None:
                #self.elaptod = tod.tod(u'0')
                if now < self.wallstart:
                    nelap = (self.wallstart - now).rawtime(0)

        if nelap != self.elapsed or dofinishtxt:
            self.elapsed = nelap
            if setarmfinish:
                self.tofinish()
                self.recalc()
                dofinishtxt = True
            if dofinishtxt:
                if type(self.meet.scbwin) is scbwin.scbtt:
                    self.meet.scbwin.setline1(u'')
                    self.meet.scbwin.setr1(u'Result:')
                    #self.meet.scbwin.setr1(u'Final Distance:')
                    if self.D is not None and self.D > 0:
                        self.meet.scbwin.sett1(u'{0:0.3f}\u2006km'.format(
                            self.D / 1000.0))
                    if self.record is not None:
                        if self.D > self.record:
                            self.meet.scbwin.setline2(u'NEW RECORD'.center(
                                self.meet.scb.linelen))
                        else:
                            self.meet.scbwin.setline2(u'')
                    else:
                        self.meet.scbwin.setline2(u'')
                    self.meet.scbwin.setr2(u'')
                    self.meet.scbwin.sett2(u'')
                    self.meet.scbwin.update()
            elif dostarttxt:
                if type(self.meet.scbwin) is scbwin.scbtt:
                    if self.record is not None:
                        self.meet.scbwin.setr1(u'Record:')
                        self.meet.scbwin.sett1(u'{0:0.3f}\u2006km'.format(
                            self.record / 1000.0))
                    elif self.target is not None:
                        self.meet.scbwin.setr1(u'Target:')
                        self.meet.scbwin.sett1(u'{0:0.3f}\u2006km'.format(
                            self.target / 1000.0))
                    if self.wallstart is not None:
                        line1 = strops.truncpad(
                            u'Start Time: ',
                            self.meet.scb.linelen - 12,
                            align='r') + strops.truncpad(
                                self.wallstart.meridiem(secs=False),
                                12,
                                align='l')
                        self.meet.scbwin.setline2(line1)

                    if nelap != u'0':
                        self.meet.scbwin.setr2(u'Countdown:')
                        self.meet.scbwin.sett2(nelap)
                    else:
                        self.meet.scbwin.setr2(u'')
                        self.meet.scbwin.sett2(u'')
                    self.meet.scbwin.update()
            else:
                self.scblap()

        # update elapsed button
        if not dostarttxt:
            self.time_lbl.set_text(self.elapsed)
        else:
            self.time_lbl.set_text(u'-' + self.elapsed)

        ## updates and running lap


# hl board,
        return True

    def armstart(self):
        """Arm timer for start trigger."""
        if self.start is None:
            _log.info(u'Arm Start.')
            self.stat_but.buttonchg(uiutil.bg_armint, u'Arm Start')
            self.meet.main_timer.arm(0)
        else:
            _log.info(u'Event already started.')

    def armsplit(self, sp, cid=4):
        """Arm timer for a 50m/200m split."""
        self.DiC = 0.0  # additional distance
        _log.info(u'armsplit')

    def abortrider(self):
        """Abort the attempt."""
        _log.debug(u'Ignored abort rider in hour event')

    def falsestart(self):
        """Register false start."""
        _log.debug(u'Ignored false start in hour event')

    def armfinish(self):
        """Arm timer for finish trigger."""
        _log.debug(u'Ignored arm finish in hour event')

    def showtimerwin(self):
        """Show timer window on scoreboard."""
        self.meet.scbwin = scbwin.scbtt(scb=self.meet.scb,
                                        header=self.event[u'pref'],
                                        subheader=self.event[u'info'])
        self.meet.scbwin.reset()
        self.recalc()

    def torunning(self):
        """Set timer running."""
        self.stat_but.buttonchg(uiutil.bg_armstart, u'Running')

    def tofinish(self):
        """Set timer finished."""
        self.stat_but.buttonchg(uiutil.bg_none, u'Finished')

    def tobell(self):
        """Set timer bell lap."""
        self.stat_but.buttonchg(uiutil.bg_armint, u'Bell')

    def tofinal(self):
        """Set timer final lap."""
        self.stat_but.buttonchg(uiutil.bg_armfin, u'Final Lap')

    def toidle(self):
        """Set timer to idle state."""
        self.finish = None
        self.start = None
        self.lstart = None
        self.lapcount = 0  # current lap counter
        self.elapsed = u''  # current elapsed time str
        #self.elaptod = tod.tod(u'0')
        self.lastlapstr = u'     '  # last lap as string
        self.laptimedirty = False
        self.splitlist = []
        self.stat_but.buttonchg(uiutil.bg_none, u'Idle')
        self.model.clear()
        sline = u'Ready'
        if self.record is not None:
            sline = u'Record: {0:0.3f}\u2006km - Ready'.format(self.record /
                                                               1000.0)
        self.infoline.set_text(sline)
        self.statusline.set_text(u'')
        _log.info(u'Reset event state to idle')

    def destroy(self):
        """Signal race shutdown."""
        self.frame.destroy()

    def show(self):
        """Show race window."""
        self.frame.show()

    def hide(self):
        """Hide race window."""
        self.frame.hide()

    def __init__(self, meet, event, ui=True):
        """Constructor."""
        self.meet = meet
        self.event = event  # Note: now a treerowref
        self.evno = event[u'evid']
        self.evtype = event[u'type']
        self.series = event[u'seri']
        self.configfile = meet.event_configfile(self.evno)
        self.autospec = ''

        self.readonly = not ui
        rstr = u''
        if self.readonly:
            rstr = u'readonly '
        self.winopen = ui
        _log.debug(u'Init %shour event %s', rstr, self.evno)
        self.comments = []

        # model
        self.reclen = tod.tod(u'1h00:00')  # record duration
        self.minlap = tod.tod(u'14.0')  # min lap time
        self.compute = u''
        self.riderstr = u''  # rider name string
        self.wallstart = None  # advertised start time
        self.finish = None  # timer finish tod
        self.start = None  # timer start tod
        self.lstart = None  # local start tod
        self.target = None  # current target in m
        self.record = None  # current record in m
        self.projlap = 12  # start showing projection after this many laps
        self.minproj = 30000  # minimum possible projection
        self.maxproj = 60000  # maximum possible projection
        self.projection = None  # current projection in m
        self.avglap = None  # current lap avg
        self.lapcount = 0  # current lap counter
        self.elapsed = u''  # current elapsed time str
        self.onestart = True  #
        self.lastlapstr = u'     '  # last lap as string
        self.laptimedirty = True
        self.splitlist = []

        # computes
        self.D = 0  # dist in m	(trunc to m)
        self.LPi = 250.0  # len of track	(should be int)
        self.TC = 0  # number of complete laps before last lap
        self.DiC = 0.0  # additional distance
        self.TTC = None  # time of last complete lap
        self.TRC = None  # time remaining to ride at beginning of
        # last lap

        # lap view model
        self.model = gtk.ListStore(gobject.TYPE_STRING, gobject.TYPE_STRING,
                                   gobject.TYPE_STRING, gobject.TYPE_STRING)

        uifile = os.path.join(metarace.UI_PATH, u'hour.ui')
        _log.debug(u'Building event interface from %r', uifile)
        b = gtk.Builder()
        b.add_from_file(uifile)

        self.frame = b.get_object(u'race_vbox')
        self.frame.connect(u'destroy', self.shutdown)

        # info pane
        self.info_expand = b.get_object(u'info_expand')
        b.get_object(u'race_info_evno').set_text(self.evno)
        self.showev = b.get_object(u'race_info_evno_show')
        self.prefix_ent = b.get_object(u'race_info_prefix')
        self.prefix_ent.connect(u'changed', self.editent_cb, u'pref')
        self.prefix_ent.set_text(self.event[u'pref'])
        self.info_ent = b.get_object(u'race_info_title')
        self.info_ent.connect(u'changed', self.editent_cb, u'info')
        self.info_ent.set_text(self.event[u'info'])

        self.time_lbl = b.get_object(u'race_info_time')
        self.time_lbl.modify_font(uiutil.DIGITFONT)

        # ctrl pane
        self.stat_but = uiutil.statbut(b.get_object(u'race_ctrl_stat_but'))
        self.stat_but.set_sensitive(True)
        self.infoline = b.get_object(u'race_ctrl_info')
        self.statusline = b.get_object(u'race_ctrl_status')
        self.statusline.modify_font(uiutil.DIGITFONT)

        # start timer and show window
        if ui:
            _log.debug(u'Connecting event ui handlers')
            # riders pane
            t = gtk.TreeView(self.model)
            self.view = t
            t.set_reorderable(False)
            t.set_enable_search(False)
            t.set_rules_hint(True)

            # riders columns
            uiutil.mkviewcoltxt(t, u'Lap  ', 0, calign=0.0)
            uiutil.mkviewcoltxt(t, u'', 1, expand=True, calign=0.0)
            uiutil.mkviewcoltxt(t, u'Split', 2, calign=1.0)
            uiutil.mkviewcoltxt(t, u'Time', 3, calign=1.0)
            t.show()
            b.get_object(u'race_result_win').add(t)
            b.connect_signals(self)



"""UNT4 Packet Wrapper."""

# Mode 1 constants
NUL = 0x00
SOH = 0x01
STX = 0x02
EOT = 0x04
HOME = 0x08
CR = 0x0d
LF = 0x0a
ERL = 0x0b
ERP = 0x0c
DLE = 0x10
DC2 = 0x12
DC3 = 0x13
DC4 = 0x14
US = 0x1f

# String encodings for Mode 1 constants over telegraph
ENCMAP = {
    unichr(NUL): u'<0>',
    unichr(SOH): u'<O>',
    unichr(STX): u'<T>',
    unichr(EOT): u'<E>',
    unichr(CR): u'<R>',
    unichr(LF): u'<A>',
    unichr(ERL): u'<L>',
    unichr(ERP): u'<P>',
    unichr(DLE): u'<D>',
    unichr(DC2): u'<2>',
    unichr(DC3): u'<3>',
    unichr(DC4): u'<4>',
    unichr(US): u'<U>',
}

# Translation map to blank forbidden control characters in header & text
TRANSMAP = {
    SOH: u' ',
    STX: u' ',
    EOT: u' ',
    ERL: u' ',
    ERP: u' ',
    DLE: u' ',
    DC2: u' ',
    DC3: u' ',
    DC4: u' ',
}


def encode(ubuf=u''):
    """Encode the unt4 buffer for use over telegraph."""
    ubuf = ubuf.replace(u'<', u'<>')
    for key in ENCMAP:
        ubuf = ubuf.replace(key, ENCMAP[key])
    return ubuf


def decode(tbuf=u''):
    """Decode the telegraph buffer to unt4 pack."""
    for key in ENCMAP:
        tbuf = tbuf.replace(ENCMAP[key], key)
    tbuf = tbuf.replace(u'<>', u'<')
    return tbuf


class unt4(object):
    """UNT4 Packet."""

    def __init__(self,
                 unt4str=None,
                 prefix=None,
                 header=u'',
                 erp=False,
                 erl=False,
                 xx=None,
                 yy=None,
                 text=u''):
        """Constructor.

        Parameters:

          unt4str -- packed unt4 string, overrides other params
          prefix -- prefix byte <DC2>, <DC3>, etc
          header -- header string eg 'R_F$'
          erp -- true for general clearing <ERP>
          erl -- true for <ERL>
          xx -- packet's column offset 0-99
          yy -- packet's row offset 0-99
          text -- packet content string

        """
        self.prefix = prefix  # <DC2>, <DC3>, etc
        self.header = header.translate(TRANSMAP)
        self.erp = erp  # true for general clearing <ERP>
        self.erl = erl  # true for <ERL>
        self.xx = xx  # input column 0-99
        self.yy = yy  # input row 0-99
        self.text = text.translate(TRANSMAP)
        if unt4str is not None:
            self.unpack(unt4str)

    def unpack(self, unt4str=u''):
        """Unpack the UNT4 unicode string into this object."""
        if len(unt4str) > 2 and unt4str[0] == unichr(SOH) \
                            and unt4str[-1] == unichr(EOT):
            self.prefix = None
            newhead = u''
            newtext = u''
            self.erl = False
            self.erp = False
            head = True  # All text before STX is considered header
            stx = False
            dle = False
            dlebuf = u''
            i = 1
            packlen = len(unt4str) - 1
            while i < packlen:
                och = ord(unt4str[i])
                if och == STX:
                    stx = True
                    head = False
                elif och == DLE and stx:
                    dle = True
                elif dle:
                    dlebuf += unt4str[i]
                    if len(dlebuf) == 4:
                        dle = False
                elif head:
                    if och in (DC2, DC3, DC4):
                        self.prefix = och  # assume pfx before head text
                    else:
                        newhead += unt4str[i]
                elif stx:
                    if och == ERL:
                        self.erl = True
                    elif och == ERP:
                        self.erp = True
                    else:
                        newtext += unt4str[i]
                i += 1
            if len(dlebuf) == 4:
                self.xx = int(dlebuf[:2])
                self.yy = int(dlebuf[2:])
            self.header = newhead
            self.text = newtext

    def pack(self):
        """Return Omega Style UNT4 unicode string packet."""
        head = u''
        text = u''
        if self.erp:  # overrides any other message content
            text = unichr(STX) + unichr(ERP)
        else:
            head = self.header
            if self.prefix is not None:
                head = unichr(self.prefix) + head
            if self.xx is not None and self.yy is not None:
                text += unichr(DLE) + u'{0:02d}{1:02d}'.format(
                    self.xx % 100, self.yy % 100)
            if self.text:
                text += self.text
            if self.erl:
                text += unichr(ERL)
            if len(text) > 0:
                text = unichr(STX) + text
        return unichr(SOH) + head + text + unichr(EOT)


# Pre-defined messages
GENERAL_CLEARING = unt4(erp=True)
GENERAL_EMPTY = unt4(xx=0, yy=0, text=u'')
OVERLAY_ON = unt4(header=u'OVERLAY ON')
OVERLAY_OFF = unt4(header=u'OVERLAY OFF')
OVERLAY_CLOCK = unt4(header=u'OVERLAY 01')
OVERLAY_MATRIX = unt4(header=u'OVERLAY 00')
OVERLAY_IMAGE = unt4(header=u'OVERLAY 02')
OVERLAY_BLANK = unt4(header=u'OVERLAY 03')
OVERLAY_GBLANK = unt4(header=u'overlay', text=u'0')
OVERLAY_GTITLE = unt4(header=u'overlay', text=u'2')
OVERLAY_GMATRIX = unt4(header=u'overlay', text=u'1')



"""HTML output library.

Cheap and nasty functional primitives for HTML output. Each primitive
returns a single string. No checking is performed on the structure of
the document produced. All elements take a named parameter 'attrs'
which is a dict of key/value attributes. Non-empty elements take a
parameter 'clist' which is a list of other constructed elements.

Note: <input> is provided by forminput()

Example for an empty element:

    hr(attrs={'id':'thehr'}) => <hr id="thehr">

Example for an element with content:

    a(['link text'], attrs={'href':'#target'}) => 

	<a href="#target">link text</a>

Example paragraph:

    p(('Check the',
       a(('website'), attrs={'href':'#website'}),
       'for more.')) => 

	<p>Check the\n<a href="#website">website</a>\nfor more.</p>

"""

from xml.sax.saxutils import escape, quoteattr
import sys


def html(headlist=(), bodylist=()):
    """Emit HTML document."""
    return u'\n'.join((preamble(), u'<html lang="en">', head(headlist),
                       body(bodylist, {'onload': 'ud();'}), u'</html>'))


def preamble():
    """Emit HTML preamble."""
    return u'<!doctype html>'


def attrlist(attrs):
    """Convert attr dict into escaped attrlist."""
    alist = []
    for a in attrs:
        alist.append(a.lower() + u'=' + quoteattr(attrs[a]))
    if len(alist) > 0:
        alist.insert(0, u'')
        return u' '.join(alist)
    else:
        return u''


def escapetext(text=u''):
    """Return escaped copy of text."""
    return escape(text, {u'"': u'&quot;'})


def comment(commentstr=u''):
    """Insert comment."""
    return u'<!-- ' + commentstr.replace(u'--', u'') + u' -->'


# output a valid but empty html templatye
def emptypage():
    return html((
        meta(attrs={u'charset': u'utf-8'}),
        meta(
            attrs={
                u'name': u'viewport',
                u'content': u'width=device-width, initial-scale=1'
            }),
        title(u'__REPORT_TITLE__'),
        link(
            attrs={
                u'href':
                u'https://cdn.jsdelivr.net/npm/bootstrap@5.0.2/dist/css/bootstrap.min.css',
                u'integrity':
                u'sha384-EVSTQN3/azprG1Anm3QDgpJLIm9Nao0Yz1ztcQTwFspd3yD65VohhpuuCOmLASjC',
                u'crossorigin': u'anonymous',
                u'rel': u'stylesheet'
            }),
        link(
            attrs={
                u'href':
                u'https://cdn.jsdelivr.net/npm/bootstrap-icons@1.8.3/font/bootstrap-icons.css',
                u'rel': u'stylesheet'
            }),
        script((
            'function ud(){null!==document.querySelector("#pgre")&&setTimeout("history.go(0);",55329)}function rl(){return setTimeout("history.go(0);",10),!1}',
        )),
    ), (u'__REPORT_NAV__',
        div((
            h1(u'__REPORT_TITLE__'),
            u'__REPORT_CONTENT__',
        ),
            attrs={u'class': u'container'})))


# Declare all the empty types
for empty in (u'meta', u'link', u'base', u'param', u'hr', u'br', u'img',
              u'col'):

    def emptyfunc(attrs={}, tag=empty):
        return u'<' + tag + attrlist(attrs) + u'>'

    setattr(sys.modules[__name__], empty, emptyfunc)


def emptyfunc(attrs={}):
    return u'<input' + attrlist(attrs) + u'>'


setattr(sys.modules[__name__], 'forminput', emptyfunc)

# Declare all the non-empties
for nonempty in (
        u'head',
        u'body',
        u'title',
        u'div',
        u'nav',
        u'header',
        u'main',
        u'style',
        u'script',
        u'p',
        u'h1',
        u'h2',
        u'h3',
        u'h4',
        u'h5',
        u'h6',
        u'ul',
        u'ol',
        u'li',
        u'dl',
        u'dt',
        u'dd',
        u'address',
        u'pre',
        u'blockquote',
        u'a',
        u'span',
        u'em',
        u'strong',
        u'dfn',
        u'code',
        u'samp',
        u'kbd',
        u'var',
        u'cite',
        u'abbr',
        u'acronym',
        u'q',
        u'sub',
        u'sup',
        u'tt',
        u'i',
        u'big',
        u'small',
        u'label',
        u'form',
        u'select',
        u'optgroup',
        u'option',
        u'textarea',
        u'fieldset',
        u'legend',
        u'button',
        u'table',
        u'caption',
        u'thead',
        u'tfoot',
        u'tbody',
        u'colgroup',
        u'tr',
        u'th',
        u'td',
):

    def nonemptyfunc(clist=(), attrs={}, elem=nonempty):
        if isinstance(clist, basestring):
            clist = (clist, )
        return (u'<' + elem + attrlist(attrs) + u'>' + u'\n'.join(clist) +
                u'</' + elem + u'>')

    setattr(sys.modules[__name__], nonempty, nonemptyfunc)



# SPDX-License-Identifier: MIT
"""Alge Timy Interface

 Interface an Alge Timy chronoprinter via serial port.

 Example:

        import metarace
        from metarace import timy
        metarace.init()

        def timercb(impulse):
            print(impulse)

        t = timy.timy()
        t.setport('/dev/ttyS0')
        t.setcb(timercb)
        t.start()
        t.sane()
        t.arm('C0')
        t.armlock()
        ...
 Configuration is read from metarace system config (metarace.json),
 under section 'timy':

  key: (type) Description [default]
  --
  baudrate: (int) serial port speed [38400]
  ctsrts: (bool) if True, use hardware flow control [False]
  chandelay: (dict) map of channel ids to delay time in seconds [{}]
             example: {'C2':'0.200'}

 Notes:

        - Callback function is only called for impulses received
          while channel is armed.

        - ALL timing impulses correctly read from an attached
          Timy will be logged by the command thread with the log
          label 'TIMER', even when the channel is not armed.

        - It is assumed that messages are received over the serial
          connection in the same order as they are measured by
          the Timy.

        - Channel delay config compensates for wireless impulse
          or relay sytems with a fixed processing time. To adjust
          channel blocking delay times, instead use method timy.delaytime()
"""

import threading
import Queue
import serial
import logging

from metarace import sysconf
from metarace import tod
from metarace.strops import chan2id, id2chan

# Configuration defaults
_DEFBAUD = 38400
_DEFCTSRTS = False

# Internal constants
_ENCODING = 'cp437'
_CHAN_UNKNOWN = -1
_CR = b'\x0d'
_TCMDS = ('EXIT', 'PORT', 'MSG', 'TRIG', 'RCLR')

# Logging
_log = logging.getLogger(u'metarace.timy')
_log.setLevel(logging.DEBUG)
_TIMER_LOG_LEVEL = 25
logging.addLevelName(_TIMER_LOG_LEVEL, u'TIMER')


def _timy_checksum(msg):
    """Return a checksum for the Timy message string."""
    ret = 0
    for ch in msg:
        ret = ret + ord(ch)
    return ret & 0xff


def _timy_getsum(chkstr):
    """Convert Timy checksum string to an integer."""
    ret = -1
    try:
        ms = (ord(chkstr[0]) - 0x30) & 0xf
        ls = (ord(chkstr[1]) - 0x30) & 0xf
        ret = ms << 4 | ls
    except Exception as e:
        _log.debug(u'error collecting timy checksum: %s', e)
    return ret


def _defcallback(impulse=None):
    """Default impulse callback."""
    _log.debug(u'Unhandled impulse: %r', impulse)


class timy(threading.Thread):
    """Timy thread object class."""

    def __init__(self):
        """Construct Timy thread object."""
        threading.Thread.__init__(self)
        self.daemon = True
        self.__port = None
        self.__cqueue = Queue.Queue()  # command queue
        self.__rdbuf = b''
        self.__arms = [
            False, False, False, False, False, False, False, False, False,
            False
        ]
        self.__clearing = False
        self.__armlocked = False
        self.__chandelay = {}
        self.__cb = _defcallback
        self.__baudrate = _DEFBAUD
        self.__ctsrts = _DEFCTSRTS
        self.name = u'timy'
        self.error = False
        if sysconf.has_option(u'timy', u'baudrate'):
            self.__baudrate = sysconf.get_posint(u'timy', u'baudrate',
                                                 _DEFBAUD)
            _log.debug(u'Set serial baudrate to: %d', self.__baudrate)
        if sysconf.has_option(u'timy', u'ctsrts'):
            self.__ctsrts = sysconf.get_bool(u'timy', u'ctsrts')
            _log.debug(u'Set serial CTSRTS to: %s', self.__ctsrts)
        if sysconf.has_option(u'timy', u'chandelay'):
            nd = sysconf.get(u'timy', u'chandelay')
            if isinstance(nd, dict):
                for cv in nd:
                    c = chan2id(cv)
                    if c != _CHAN_UNKNOWN:
                        nv = tod.mktod(nd[cv])
                        self.__chandelay[c] = nv
                        _log.debug(u'Set channel delay %s: %s', c,
                                   nv.rawtime(4))
            else:
                _log.debug(u'Invalid channel delay setting: %r', nd)

    def setcb(self, func=None):
        """Set or clear the event callback."""
        if func is not None:
            self.__cb = func
        else:
            self.__cb = _defcallback

    def printline(self, msg=u''):
        """Print msg to Timy printer, stripped and truncated."""
        lmsg = msg[0:32]
        _log.log(_TIMER_LOG_LEVEL, lmsg)
        self.__cqueue.put_nowait((u'MSG', u'DTP' + lmsg + u'\r'))

    def linefeed(self):
        """Advance Timy printer by one line."""
        self.__cqueue.put_nowait((u'MSG', u'PRILF\r'))

    def clrmem(self):
        """Clear memory in attached Timy."""
        self.__cqueue.put_nowait((u'MSG', u'CLR\r'))

    def status(self):
        """Request status and current program."""
        self.__cqueue.put_nowait((u'MSG', u'NSF?\r'))
        self.__cqueue.put_nowait((u'MSG', u'PROG?\r'))

    def dumpall(self):
        """Request a dump of all times to host."""
        self.__cqueue.put_nowait((u'MSG', u'RSM\r'))

    def delaytime(self, newdelay=u'2.0'):
        """Update blocking delay time for all channels."""
        dt = tod.mktod(newdelay)
        if dt is not None:
            if dt > tod.ZERO and dt < tod.tod(u'99.99'):
                nt = dt.rawtime(2, zeros=True)[6:]
                self.__cqueue.put_nowait((u'MSG', u'DTS' + nt + u'\r'))
                self.__cqueue.put_nowait((u'MSG', u'DTF' + nt + u'\r'))
            else:
                _log.info(u'Ignoring invalid delay time: %s', dt.rawtime())
        else:
            _log.info(u'Ignoring invalid delay time')

    def printer(self, enable=False):
        """Enable or disable printer."""
        cmd = u'0'
        if enable:
            cmd = u'1'
        self.__cqueue.put_nowait((u'MSG', u'PRINTER' + cmd + u'\r'))

    def printimp(self, doprint=True):
        """Enable or disable internal print of timing impulses."""
        cmd = u'1'
        if doprint:
            cmd = u'0'
        self.__cqueue.put_nowait((u'MSG', u'PRIIGN' + cmd + u'\r'))

    def keylock(self, setlock=True):
        """Set or clear the timy keypad lock function."""
        cmd = u'1'
        if not setlock:
            cmd = u'0'
        self.__cqueue.put_nowait((u'MSG', u'KL' + cmd + u'\r'))

    def write(self, msg=None):
        """Queue a raw command string to attached Timy."""
        self.__cqueue.put_nowait((u'MSG', msg.rstrip() + u'\r'))

    def exit(self, msg=None):
        """Request thread termination."""
        self.running = False
        self.__cqueue.put_nowait((u'EXIT', msg))

    def setport(self, device=None):
        """Request (re)opening serial port as specified.

        Call setport with None or an empty string to close an open port
        or to run the Timy thread with no external device.

        """
        self.__cqueue.put_nowait((u'PORT', device))

    def arm(self, channel=0):
        """Arm channel 0 - 8 for timing impulses."""
        chan = chan2id(channel)
        _log.debug(u'Arming channel %s', id2chan(chan))
        self.__arms[chan] = True

    def dearm(self, channel=0):
        """Disarm channel 0 - 8."""
        chan = chan2id(channel)
        _log.debug(u'De-arm channel %s', id2chan(chan))
        self.__arms[chan] = False

    def armlock(self, lock=True):
        """Set or clear the arming lock."""
        self.__armlocked = bool(lock)
        _log.debug(u'Armlock is now %s', self.__armlocked)

    def sane(self):
        """Initialise Timy to 'sane' values.

        Values set by sane():

            TIMIYINIT		- initialise
            KL0			- keylock off
	    CHK1		- enable "checksum"
	    PRE4		- 10,000th sec precision
	    RR0			- Round by 'cut'
	    BE1			- Beep on
	    DTS02.00		- Start delay 2.0
	    DTF02.00		- Finish & intermediate delay 2.0
	    EMU0		- Running time off
	    PRINTER0		- Printer off
	    PRIIGN1		- Don't print impulses to receipt
            SL0			- Logo off
	    PRILF		- Linefeed
	
        All commands are queued individually to the command thread
        so it may be necessary to use wait() to suspend the calling
        thread until all the commands are sent:

            t.start()
	    t.sane()
	    t.wait()
    
        Note: "sane" here comes from use at track meets with the
              metarace program. It may not always make sense eg, to
              have all channel delays set to 2 hundredths of a
              second, or to have the internal impulse print off
              by default.

        """
        for msg in [
                u'TIMYINIT', u'NSF?', u'PROG?', u'KL0', u'CHK1', u'PRE4',
                u'RR0', u'BE1', u'DTS02.00', u'DTF02.00', u'EMU0', u'PRINTER0',
                u'PRIIGN1', u'SL0', u'PRILF'
        ]:
            self.write(msg)

    def wait(self):
        """Wait for Timy thread to finish processing any queued commands."""
        self.__cqueue.join()

    def __parse_message(self, msg):
        """Return tod object from timing msg or None."""
        ret = None
        msg = msg.rstrip()  # remove cr/lf if present
        tsum = 0
        csum = 0
        if len(msg) == 28:
            # assume checksum present, grab it and truncate msg
            tsum = _timy_getsum(msg[26:28])
            msg = msg[0:26]
            csum = _timy_checksum(msg)
        if len(msg) == 26:
            # assume now msg is a timing impulse
            if tsum == csum:
                e = msg.split()
                if len(e) == 4:
                    cid = chan2id(e[1])
                    ret = tod.mktod(e[2])
                    if ret is not None:
                        if cid in self.__chandelay:
                            # note: ret might wrap over 24hr boundary
                            ret = ret - self.__chandelay[cid]
                        ret.index = e[0]
                        ret.chan = e[1]
                        ret.refid = u''
                        ret.source = self.name
                    else:
                        _log.error(u'Invalid message: %s', msg)
                else:
                    _log.error(u'Invalid message: %s', msg)
            else:
                _log.error(u'Corrupt message: %s', msg)
                _log.error(u'Checksum fail: 0x%02X != 0x%02X', tsum, csum)
        else:
            msg = msg.strip()
            if msg == u'CLR':
                self.__cqueue.put_nowait((u'RCLR', u''))
        return ret

    def __proc_impulse(self, st):
        """Process a parsed tod impulse from the Timy.

        On reception of a timing channel message, the channel is
        compared against the list of armed channels. If the channel
        is armed, the callback is run.

        If arm lock is not set, the channel is then de-armed.
        """
        _log.log(_TIMER_LOG_LEVEL, st)
        channo = chan2id(st.chan)
        if channo != _CHAN_UNKNOWN:
            if self.__arms[channo]:
                self.__cb(st)
                if not self.__armlocked:
                    self.__arms[channo] = False
            if st.index.isdigit():
                index = int(st.index)
                if index > 2000 and not self.__clearing:
                    self.__clearing = True
                    self.clrmem()
                    _log.debug(u'Auto clear memory')
        else:
            pass
        return False

    def __read(self):
        """Read messages from timy until a timeout condition."""
        ch = self.__port.read(1)
        mcnt = 0
        while ch != b'':
            if ch == _CR:
                # Return ends the current 'message'
                self.__rdbuf += ch  # include trailing <cr>
                msg = self.__rdbuf.decode(_ENCODING, 'ignore')
                _log.debug(u'RECV: %r', msg)
                t = self.__parse_message(msg)
                if t is not None:
                    self.__proc_impulse(t)
                self.__rdbuf = b''
                mcnt += 1
                if mcnt > 4:  # break to allow write back
                    return
            else:
                self.__rdbuf += ch
            ch = self.__port.read(1)

    def run(self):
        """Run the Timy thread.

           Called by invoking thread method: timy.start()
        """
        running = True
        _log.debug(u'Starting')
        while running:
            try:
                # Read phase
                if self.__port is not None:
                    self.__read()
                    m = self.__cqueue.get_nowait()
                else:
                    m = self.__cqueue.get()
                self.__cqueue.task_done()

                # Write phase
                if isinstance(m, tuple) and m[0] in _TCMDS:
                    if m[0] == u'MSG':
                        if self.__port is not None and not self.error:
                            _log.debug(u'SEND: %r', m[1])
                            self.__port.write(m[1].encode(_ENCODING, 'ignore'))
                    elif m[0] == u'TRIG':
                        if isinstance(m[1], tod.tod):
                            self.__proc_impulse(m[1])
                    elif m[0] == u'RCLR':
                        self.__clearing = False
                    elif m[0] == u'EXIT':
                        _log.debug(u'Request to close: %s', m[1])
                        running = False
                    elif m[0] == u'PORT':
                        if self.__port is not None:
                            self.__port.close()
                            self.__port = None
                        if m[1] is not None and m[1] not in [
                                u'', u'NULL', u'None'
                        ]:
                            _log.debug(u'Re-Connect port: %s @ %d', m[1],
                                       self.__baudrate)
                            self.__port = serial.Serial(m[1],
                                                        self.__baudrate,
                                                        rtscts=self.__ctsrts,
                                                        timeout=0.2)
                            self.error = False
                        else:
                            _log.debug(u'Not connected')
                            self.error = True
                    else:
                        pass
                else:
                    _log.warning(u'Unknown message: %r', m)
            except Queue.Empty:
                pass
            except serial.SerialException as e:
                if self.__port is not None:
                    self.__port.close()
                    self.__port = None
                self.error = True
                _log.error(u'Serial error: %s', e)
            except Exception as e:
                _log.error(u'%s: %s', e.__class__.__name__, e)
                self.error = True
        if self.__port is not None:
            self.__port.close()
            self.__port = None
        _log.info(u'Exiting')



# SPDX-License-Identifier: MIT
"""String filtering, truncation and padding."""

# Note: These functions consider unicode string length and
#       displayed string length to be equal, so any string with zero
#	length characters (eg combining) will be incorrectly
#	truncated and/or padded. Output to fixed-width displays
#	like DHI and track announce will be incorrect.

import re
from random import randint

# replace codepoints 0->255 with space unless overridden
# "protective" against unencoded ascii strings and control chars
SPACEBLOCK = u''
for i in xrange(0, 256):
    SPACEBLOCK += unichr(i)


# unicode translation 'map' class
class unicodetrans(object):

    def __init__(self, keep=u'', replace=SPACEBLOCK, replacechar=u' '):
        self.comp = dict((ord(c), replacechar) for c in replace)
        for c in keep:
            self.comp[ord(c)] = c

    def __getitem__(self, k):  # override to return a None
        return self.comp.get(k)


INTEGER_UTRANS = unicodetrans(u'-0123456789')
NUMERIC_UTRANS = unicodetrans(u'-0123456789.e')
PLACELIST_UTRANS = unicodetrans(
    u'-0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz')
PLACESERLIST_UTRANS = unicodetrans(
    u'-.0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz')
BIBLIST_UTRANS = unicodetrans(
    u'0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz')
RIDERNO_UTRANS = unicodetrans(
    u'0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz', u'',
    u'')
BIBSERLIST_UTRANS = unicodetrans(
    u'.0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz')
WEBFILE_UTRANS = unicodetrans(
    u'_0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz', u'.',
    u'_')
# special case: map controls and spaces, but keep everything else
PRINT_UTRANS = {}
for cp in xrange(0, 0x20):
    PRINT_UTRANS[cp] = u' '
for cp in xrange(0x7f, 0xa1):
    PRINT_UTRANS[cp] = u' '
PRINT_UTRANS[0x1680] = u' '
PRINT_UTRANS[0x180e] = u' '
PRINT_UTRANS[0x202f] = u' '
PRINT_UTRANS[0x205f] = u' '
PRINT_UTRANS[0x3000] = u' '
PRINT_UTRANS[0xffa0] = u' '

# timing channels - this duplicates defs in timy
CHAN_START = 0
CHAN_INT = 9
CHAN_UNKNOWN = -1

# running number comparisons
RUNNER_NOS = {
    u'red': 0,
    u'whi': 1,
    u'blu': 2,
    u'yel': 3,
    u'grn': 4,
    u'pin': 5,
    u'bla': 6,
    u'gry': 7,
    u'ora': 8,
    u'pur': 9,
    u'rdw': 10,
    u'blw': 11,
    u'ylw': 12,
    u'grw': 13
}

DNFCODEMAP = {u'otl': 1, u'dsq': 4, u'dnf': 3, u'dns': 5, u'': 2}


def cmp_dnf(x, y):
    """Comparison func for two dnf codes."""
    if x not in DNFCODEMAP:
        x = u''
    if y not in DNFCODEMAP:
        y = u''
    return cmp(DNFCODEMAP[x], DNFCODEMAP[y])


def rand_key(data=None):
    """Return a random integer key for shuffling."""
    return randint(0, 0xffffffff)


def riderno_key(bib):
    """Return a comparison key for sorting rider number strings."""
    return bibstr_key(bib)


def dnfcode_key(code):
    """Return a rank/dnf code sorting key."""
    # rank [rel] '' dsq otl dnf dns
    dnfordmap = {
        u'rel': 8000,
        u'': 8500,
        u'otl': 8800,
        u'dnf': 9000,
        u'dns': 9500,
        u'dsq': 10000,
    }
    ret = 0
    if code is not None:
        code = code.lower()
        if code in dnfordmap:
            ret = dnfordmap[code]
        else:
            code = code.strip(u'.')
            if code.isdigit():
                ret = int(code)
    return ret


def bibstr_key(bibstr=u''):
    """Return a comparison key for sorting rider bib.ser strings."""
    (bib, ser) = bibstr2bibser(bibstr)
    bval = 0
    if bib.isdigit():
        bval = int(bib)
    else:
        sbib = bib.translate(INTEGER_UTRANS).strip()
        if sbib and sbib.isdigit():
            bval = int(sbib)
        else:
            if bib.lower()[0:3] in RUNNER_NOS:
                bval = RUNNER_NOS[bib.lower()[0:3]]
            else:
                bval = id(bib)
    sval = 0
    if ser != u'':
        sval = ord(ser[0]) << 12
    return sval | (bval & 0xfff)


def randstr(data=None):
    """Return a string of 6 random digits."""
    return unicode(randint(10000, 99999))


def promptstr(prompt=u'', value=u''):
    """Prefix a non-empty string with a prompt, or return empty."""
    ret = u''
    if value:
        ret = prompt + u' ' + value
    return ret


def listsplit(liststr=u''):
    """Return a split and stripped list."""
    ret = []
    for e in liststr.split(u','):
        ret.append(e.strip())
    return ret


def heatsplit(heatstr):
    """Return a failsafe heat/lane pair for the supplied heat string."""
    hv = heatstr.split(u'.')
    while len(hv) < 2:
        hv.append(u'0')
    return (riderno_key(hv[0]), riderno_key(hv[1]))


def fitname(first, last, width, trunc=False):
    """Return a truncated name field for fixed-width display.

    Attempts to modify name to fit within width as follows:

    'First Lastone-Lasttwo'

	1: Split lastname ->	'First Lasttwo'
	2: Abbreviate first -> 	'F. Lastone-Lasttwo'
	3: Both ->		'F. Lasttwo'

    If trunc is set, truncate and replace final char with
    ellipsis u'\u2026':		'F. Lastt...'
    """
    ret = u''
    fstr = first.strip().title()
    lstr = last.strip().upper()
    if len(fstr) + len(lstr) >= width:
        lshrt = lstr.split(u'-')[-1].strip()
        if len(fstr) + len(lshrt) >= width:
            fshrt = fstr
            if fstr and len(fstr) > 1:
                fshrt = fstr[0] + u'.'
            if len(fshrt) + len(lstr) >= width:
                lstr = lshrt
                if len(fstr) + len(lshrt) >= width:
                    # Abbrev first with split last
                    fstr = fshrt
                else:
                    # Full first with split last
                    pass
            else:
                # Abbrev first with full last
                fstr = fshrt
        else:
            # Full first with split last
            lstr = lshrt
    ret = u' '.join([fstr, lstr]).strip()
    if trunc and len(ret) > width:
        if width > 4:
            ret = ret[0:(width - 1)] + u'\u2026'  # Ellipsis
        else:
            ret = ret[0:width]
    return ret


def drawno_encirc(drawstr=u''):
    ret = drawstr
    if drawstr.isdigit():
        try:
            ival = int(drawstr)
            if ival > 0 and ival <= 10:
                ret = (
                    u'\u00a0' +  # nbsp to get full line height
                    unichr(0x245f + ival))  # CP U+2460 "Circled digit"
        except Exception:
            pass
    return ret


def rank2ord(place):
    """Return ordinal for the given place."""
    omap = {
        u'1': u'st',
        u'2': u'nd',
        u'3': u'rd',
        u'11': u'th',
        u'12': u'th',
        u'13': u'th'
    }
    ret = place
    if place.isdigit():
        if place in omap:
            ret = place + omap[place]
        elif len(place) > 1 and place[-2:] in omap:
            ret = place + omap[place[-2:]]
        else:
            if len(place) > 1 and place[-1] in omap:  # last digit 1,2,3
                ret = place + omap[place[-1]]
            else:
                ret = place + u'th'
    return ret


def rank2int(rank):
    """Convert a rank/placing string into an integer."""
    ret = None
    try:
        ret = int(rank.replace(u'.', u''))
    except Exception:
        pass
    return ret


def mark2int(handicap):
    """Convert a handicap string into an integer number of metres."""
    handicap = handicap.strip().lower()
    ret = None  # not recognised as handicap
    if handicap != u'':
        if handicap[0:3] == u'scr':  # 'scr{atch}'
            ret = 0
        else:  # try [number]m form
            handicap = handicap.translate(INTEGER_UTRANS).strip()
            try:
                ret = int(handicap)
            except Exception:
                pass
    return ret


def truncpad(srcline, length, align=u'l', ellipsis=True):
    """Return srcline truncated and padded to length, aligned as requested."""
    # truncate
    if len(srcline) > length:
        if ellipsis and length > 4:
            ret = srcline[0:(length - 1)] + u'\u2026'  # Ellipsis
        else:
            ret = srcline[0:length]
    else:
        # pad
        if len(srcline) < length:
            if align == u'l':
                ret = srcline.ljust(length)
            elif align == u'r':
                ret = srcline.rjust(length)
            else:
                ret = srcline.center(length)
        else:
            ret = srcline
    return ret


def resname_bib(bib, first, last, club):
    """Return rider name formatted for results with bib."""
    ret = [bib, u' ', fitname(first, last, 64)]
    if club is not None and club != u'':
        if len(club) < 4:
            club = club.upper()
        ret.extend([u' (', club, u')'])
    return u''.join(ret)


def resname(first, last=None, club=None):
    """Return rider name formatted for results."""
    ret = fitname(first, last, 64)
    if club is not None and club != u'':
        if len(club) < 4:
            club = club.upper()
        ret = u''.join([ret, u' (', club, u')'])
    return ret


def listname(first, last=None, club=None, maxlen=32):
    """Return a rider name summary field for non-edit lists."""
    ret = fitname(first, last, maxlen)
    if club:
        if len(club) < 4:
            club = club.upper()
        ret = u''.join([ret, u' (', club, u')'])
    return ret


def reformat_bibserlist(bibserstr):
    """Filter and return a bib.ser start list."""
    return u' '.join(bibserstr.translate(BIBSERLIST_UTRANS).split())


def reformat_bibserplacelist(placestr):
    """Filter and return a canonically formatted bib.ser place list."""
    if u'-' not in placestr:  # This is the 'normal' case!
        return reformat_bibserlist(placestr)
    # otherwise, do the hard substitutions...
    placestr = placestr.translate(PLACESERLIST_UTRANS).strip()
    placestr = re.sub(r'\s*\-\s*', r'-', placestr)  # remove surrounds
    placestr = re.sub(r'\-+', r'-', placestr)  # combine dupes
    return u' '.join(placestr.strip(u'-').split())


def reformat_biblist(bibstr):
    """Filter and return a canonically formatted start list."""
    return u' '.join(bibstr.translate(BIBLIST_UTRANS).split())


## TODO: remove def reformat_riderlist(riderstr, rdb=None, series=u''):


def riderlist_split(riderstr, rdb=None, series=u''):
    """Filter, search and return a list of matching riders for entry."""
    ret = []
    riderstr = riderstr.lower()

    # special case: 'all' -> return all riders from the specified series.
    if rdb is not None and riderstr.strip() == u'all':
        riderstr = u''
        for r in rdb:
            if r[5] == series:
                ret.append(r[0])

    # pass 1: search for categories
    if rdb is not None:
        for cat in sorted(rdb.listcats(series), key=len, reverse=True):
            if len(cat) > 0 and cat.lower() in riderstr:
                ret.extend(rdb.biblistfromcat(cat, series))
                riderstr = riderstr.replace(cat.lower(), u'')

    # pass 2: append riders and expand any series if possible
    riderstr = reformat_placelist(riderstr)
    for nr in riderstr.split():
        if u'-' in nr:
            # try for a range...
            l = None
            n = None
            for r in nr.split(u'-'):
                if l is not None:
                    if l.isdigit() and r.isdigit():
                        start = int(l)
                        end = int(r)
                        if start < end:
                            c = start
                            while c < end:
                                ret.append(unicode(c))
                                c += 1
                        else:
                            ret.append(l)
                    else:
                        # one or both not ints
                        ret.append(l)
                else:
                    pass
                l = r
            if l is not None:  # catch final value
                ret.append(l)
        else:
            ret.append(nr)
    return ret


def placeset(spec=u''):
    """Convert a place spec into an ordered set of place ints."""

    # NOTE: ordering of the set must be retained to correctly handle
    #       autospecs where the order of the places is not increasing
    #       eg: sprint semi -> sprint final, the auto spec is: 3,1,2,4
    #       so the 'winners' go to the gold final and the losers to the
    #       bronze final.
    ret = []
    spec = reformat_placelist(spec)
    # pass 1: expand ranges
    for nr in spec.split():
        if u'-' in spec:
            # try for a range...
            l = None
            n = None
            for r in nr.split(u'-'):
                if l is not None:
                    if l.isdigit() and r.isdigit():
                        start = int(l)
                        end = int(r)
                        if start < end:
                            c = start
                            while c < end:
                                ret.append(unicode(c))
                                c += 1
                        else:
                            ret.append(l)  # give up on last val
                    else:
                        # one or both not ints
                        ret.append(l)
                else:
                    pass
                l = r
            if l is not None:  # catch final value
                ret.append(l)
        else:
            ret.append(nr)
    # pass 2: filter out non-numbers, only places considered
    rset = []
    for i in ret:
        if i.isdigit():
            ival = int(i)
            if ival not in rset:
                rset.append(ival)
    return rset


def reformat_placelist(placestr):
    """Filter and return a canonically formatted place list."""
    if u'-' not in placestr:
        return reformat_biblist(placestr)
    # otherwise, do the hard substitutions...
    placestr = placestr.translate(PLACELIST_UTRANS).strip()
    placestr = re.sub(r'\s*\-\s*', r'-', placestr)  # remove surrounds
    placestr = re.sub(r'\-+', r'-', placestr)  # combine dupes
    return u' '.join(placestr.strip(u'-').split())


def confopt_str(confob, default=None):
    """Check and return a plain string for the provided value."""
    ret = default
    if isinstance(confob, basestring):
        ret = confob
    return ret


def confopt_bool(confstr):
    """Check and return a boolean option from config."""
    if isinstance(confstr, basestring):
        if confstr.lower() in [u'yes', u'true', u'1']:
            return True
        else:
            return False
    else:
        return bool(confstr)


def plural(count=0):
    """Return plural extension for provided count."""
    ret = u's'
    if count == 1:
        ret = u''
    return ret


def confopt_riderno(confstr, default=u''):
    """Check and return rider number, filtered only."""
    return confstr.translate(RIDERNO_UTRANS).strip()


def confopt_float(confstr, default=None):
    """Check and return a floating point number."""
    ret = default
    try:
        ret = float(confstr)
    except Exception:
        pass
    return ret


def confopt_distunits(confstr):
    """Check and return a valid unit from metres or laps."""
    if u'lap' in confstr.lower():
        return u'laps'
    else:
        return u'metres'


def confopt_int(confstr, default=None):
    """Check and return a valid integer."""
    ret = default
    try:
        ret = int(confstr)
    except Exception:
        pass
    return ret


def confopt_posint(confstr, default=None):
    """Check and return a valid positive integer."""
    ret = default
    try:
        ret = int(confstr)
        if ret < 0:
            ret = default
    except Exception:
        pass
    return ret


def confopt_dist(confstr, default=None):
    """Check and return a valid distance unit."""
    return confopt_posint(confstr, default)


def chan2id(chanstr=u'0'):
    """Return a channel ID for the provided string, without fail."""
    ret = CHAN_UNKNOWN
    try:
        if isinstance(chanstr, basestring):
            chanstr = chanstr.upper().rstrip(u'M').lstrip(u'C')
            if chanstr.isdigit():
                ret = int(chanstr)
        else:
            ret = int(chanstr)
    except Exception as e:
        pass
    if ret < CHAN_UNKNOWN or ret > CHAN_INT:
        ret = CHAN_UNKNOWN
    return ret


def id2chan(chanid=0):
    """Return a normalised channel string for the provided channel id."""
    ret = u'C?'
    if isinstance(chanid, int) and chanid >= CHAN_START and chanid <= CHAN_INT:
        ret = u'C' + unicode(chanid)
    return ret


def confopt_chan(confstr, default=None):
    """Check and return a valid timing channel id string."""
    ret = chan2id(confstr)
    if ret == CHAN_UNKNOWN:
        ret = chan2id(default)
    return ret


def confopt_pair(confstr, value, default=None):
    """Return value or the default."""
    ret = default
    if confstr.lower() == value.lower():
        ret = value
    return ret


def confopt_list(confstr, list=[], default=None):
    """Return an element from list or default."""
    ret = default
    look = confstr.lower()
    if look in list:
        ret = look
    return ret


def bibstr2bibser(bibstr=u''):
    """Split a bib.series string and return bib and series."""
    a = bibstr.strip().split(u'.')
    ret_bib = u''
    ret_ser = u''
    if len(a) > 0:
        ret_bib = a[0]
    if len(a) > 1:
        ret_ser = a[1]
    return (ret_bib, ret_ser)


def lapstring(lapcount=None):
    lapstr = u''
    if lapcount:
        lapstr = unicode(lapcount) + u' Lap'
        if lapcount > 1:
            lapstr += u's'
    return lapstr


def bibser2bibstr(bib=u'', ser=u''):
    """Return a valid bib.series string."""
    ret = bib
    if ser != u'':
        ret += u'.' + ser
    return ret


def titlesplit(src=u'', linelen=24):
    """Split a string on word boundaries to try and fit into 3 fixed lines."""
    ret = [u'', u'', u'']
    words = src.split()
    wlen = len(words)
    if wlen > 0:
        line = 0
        ret[line] = words.pop(0)
        for word in words:
            pos = len(ret[line])
            if pos + len(word) >= linelen:
                # new line
                line += 1
                if line > 2:
                    break
                ret[line] = word
            else:
                ret[line] += u' ' + word
    return ret



# SPDX-License-Identifier: MIT
"""Race Result System decoder interface."""

from __future__ import division

import Queue
import logging
import socket
import datetime

from . import (decoder, DECODER_LOG_LEVEL)
from metarace import tod
from metarace import sysconf

_log = logging.getLogger(u'metarace.decoder.rrs')
_log.setLevel(logging.DEBUG)

# desired target protocol level
_RRS_PROTOCOL = u'3.2'
# decoder port
_RRS_TCP_PORT = 3601
# passing record length
_RRS_PASSLEN = 20
# status record length
_RRS_STATUSLEN = 25
# Warn if battery voltage is below this many volts
_RRS_LOWBATT = 2.0
_RRS_SYNFMT = u'SETTIME;{:04d}-{:02d}-{:02d};{}'
_RRS_EOL = u'\r\n'
_RRS_MARKER = u'99999'
_RRS_ENCODING = u'iso8859-1'
_RRS_IOTIMEOUT = 1.0
_RRS_PASSIVEID = u'1'

# Error flags from decoder status
_RRS_ERRORFLAGS = {
    1: u'UHF module reports an error',
    16: u'Active loop error',
    32: u'Active loop limit',
    64: u'Active connection lost',
    256: u'GPS time sync error',
    512: u'GPS communication error warning',
    1024: u'Active time sync error'
}


class rrs(decoder):
    """RRS thread object class."""

    def __init__(self):
        decoder.__init__(self)
        self._io = None
        self._rdbuf = b''
        self._curfile = None
        self._lastpassing = None
        self._dorefetch = True
        self._fetchpending = False
        self._pending_command = None
        self._allowstored = False
        self._passiveloop = _RRS_PASSIVEID

    # API overrides
    def sync(self, data=None):
        self.stop_session(data)
        self._cqueue.put_nowait((u'_sync', data))
        self.start_session(data)

    def start_session(self, data=None):
        self.write(u'STARTOPERATION')
        self.write(u'PASSINGS')

    def stop_session(self, data=None):
        self.write(u'STOPOPERATION')

    def status(self, data=None):
        self.write(u'GETSTATUS')

    def connected(self):
        return self._io is not None

    def clear(self, data=None):
        self.stop_session(data)
        self.write(u'CLEARFILES')
        self._cqueue.put_nowait((u'_sync', data))
        self.start_session(data)

    # Device-specific functions
    def _close(self):
        if self._io is not None:
            _log.debug(u'Close connection')
            cp = self._io
            self._io = None
            try:
                cp.shutdown(socket.SHUT_RDWR)
            except Exception as e:
                _log.debug(u'%s: shutdown socket: %s', e.__class__.__name__, e)
            cp.close()

    def _sane(self, data=None):
        for m in [
                u'GETPROTOCOL',
                u'SETPROTOCOL;{}'.format(_RRS_PROTOCOL),
                u'SETPUSHPREWARNS;0',
                u'SETPUSHPASSINGS;1;0',
                u'GETCONFIG;GENERAL;BOXNAME',
                u'GETCONFIG;ACTIVE;LOOPID',
                u'GETCONFIG;ACTIVE;POWER',
                u'GETCONFIG;UPLOAD;CONNECTION',
                u'GETINTERFACES',
                u'GETSTATUS',
                u'PASSINGS',
        ]:
            self.write(m)

    def _port(self, port):
        """Re-establish connection to supplied device port."""
        self._close()
        addr = (port, _RRS_TCP_PORT)
        _log.debug(u'Connecting to %r', addr)
        self._rdbuf = b''
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            s.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
            s.setsockopt(socket.IPPROTO_TCP, socket.TCP_USER_TIMEOUT, 5000)
        except Exception as e:
            _log.debug('%s setting socket option: %s', e.__class__.__name__, e)
        s.connect(addr)
        s.settimeout(_RRS_IOTIMEOUT)
        self._io = s
        self.sane()

    def _sync(self, data=None):
        a = datetime.datetime.now()
        nt = tod.now() + tod.tod('0.640')
        syncmd = _RRS_SYNFMT.format(
            a.year, a.month, a.day,
            nt.rawtime(places=3, zeros=True, hoursep=u':'))
        self._write(syncmd)

    def _replay(self, file):
        """RRS-specific replay"""
        if file.isdigit():
            _log.debug(u'Replay passings from %r', file)
            cmd = u'GETFILE;{:d}'.format(file)
            self._write(cmd)
        else:
            _log.error(u'Invalid file specified: %r', file)

    def _write(self, msg):
        if self._io is not None:
            ob = (msg + _RRS_EOL)
            self._io.sendall(ob.encode(_RRS_ENCODING))
            #_log.debug(u'SEND: %r', ob)

    def _passing(self, pv):
        """Process a RRS protocol passing."""
        if len(pv) == _RRS_PASSLEN:
            today = datetime.date.today().isoformat()
            istr = pv[0]  # <PassingNo>
            tagid = pv[1]  # <Bib/TranspCode>
            date = pv[2]  # <Date>
            timestr = pv[3]  # <Time>
            eventid = pv[4]  # <EventID>
            isactive = pv[8]  # <IsActive>
            loopid = pv[10]  # <LoopID>
            wuc = pv[12]  # <WakeupCounter>
            battery = pv[13]  # <Battery>
            adata = pv[15]  # <InternalActiveData>
            bname = pv[16]  # <BoxName>
            hits = pv[5]  # <Hits>
            rssi = pv[6]  # <MaxRSSI>

            # An error here will invalidate the whole passing
            pid = int(istr)
            if self._lastpassing is not None:
                expectpid = self._lastpassing + 1
                if pid != expectpid:
                    _log.debug(u'Ignore out of sequence passing: %r != %r',
                               pid, expectpid)
                    return
            self._lastpassing = pid

            active = False
            try:
                active = bool(int(isactive))
                if eventid == u'0':
                    eventid = u''
            except Exception as e:
                _log.debug(u'%s reading isactive: %s', e.__class__.__name__, e)

            if not loopid:
                loopid = self._passiveloop
            try:
                loopid = u'C' + str(int(loopid))
            except Exception as e:
                _log.debug(u'%s reading loop id: %s', e.__class__.__name__, e)
            activestore = False
            if active and adata:
                activestore = (int(adata) & 0x40) == 0x40
            if not active and tagid and eventid:
                tagid = u'-'.join((eventid, tagid))
            #if tagid.startswith(_RRS_MARKER):
            if tagid == _RRS_MARKER:
                tagid = u''

            if battery and tagid:
                try:
                    bv = float(battery)
                    if bv < _RRS_LOWBATT:
                        _log.warning(u'Low battery %s: %0.1fV', tagid, bv)
                except Exception as e:
                    _log.debug(u'%s reading battery voltage: %s',
                               e.__class__.__name__, e)

            if hits and rssi and tagid:
                try:
                    hitcount = int(hits)
                    if active:
                        rssival = int(rssi, 16)
                        twofour = -90 + ((rssival & 0x70) >> 2)
                        lstrength = 1 + (rssival & 0x0f)
                        if lstrength < 5 or twofour < -82 or hitcount < 4:
                            _log.warning(
                                u'Poor read %s: Hits:%d RSSI:%ddBm Loop:%ddB',
                                tagid, hitcount, twofour, lstrength)
                    else:
                        rssival = int(rssi)
                        if rssival < -82 or hitcount < 4:
                            _log.warning(u'Poor read %s: Hits:%d RSSI:%ddBm',
                                         tagid, hitcount, rssival)
                except Exception as e:
                    _log.debug(u'%s reading hits/RSSI: %s',
                               e.__class__.__name__, e)

            # emit a decoder log line TBD
            _log.log(DECODER_LOG_LEVEL, u';'.join(pv))

            # accept valid passings and trigger callback
            t = tod.mktod(timestr)
            if t is not None:
                t.index = istr
                t.chan = loopid
                t.refid = tagid
                t.source = bname
                if not activestore or self._allowstored:
                    self._trig(t)
                else:
                    pass
        else:
            _log.info(u'Non-passing message: %r', pv)

    def _statusmsg(self, pv):
        """Process a RRS protocol status message."""
        if len(pv) == _RRS_STATUSLEN:
            pwr = pv[2]  # <HasPower>
            opmode = pv[4]  # <IsInOperationMode>
            uhf = pv[8]  # <ReaderIsHealthy>
            batt = pv[9]  # <BatteryCharge>
            active = pv[13]  # <ActiveExtConnected>
            eflag = pv[23]  # <ErrorFlags>
            if batt == u'-1':
                batt = u'[estimating]'
            else:
                batt += u'%'
            loopch = u'n/a'
            loopid = self._passiveloop
            looppower = u'n/a'
            if active == u'1':
                loopch = pv[14]
                loopid = pv[15]
                looppower = pv[16]
            else:
                loopch = pv[3]

            if opmode == u'1':
                _log.info(
                    u'Started, charge:%s, uhf:%s, batt:%s, ch:%s, loop:%s, power:%s',
                    pwr, uhf, batt, loopch, loopid, looppower)
            else:
                _log.warning(
                    u'Not started, charge:%s, uhf:%s, batt:%s, ch:%s, loop:%s, power:%s',
                    pwr, uhf, batt, loopch, loopid, looppower)

            if eflag != u'0':
                stat = int(eflag)
                evec = []
                for bit in _RRS_ERRORFLAGS:
                    if stat & bit == bit:
                        evec.append(_RRS_ERRORFLAGS[bit])
                _log.error(u'Error: %s', u', '.join(evec))

    def _configmsg(self, pv):
        """Process a RRS protocol config message."""
        if len(pv) > 3:
            if pv[1] == u'BOXNAME':
                boxid = pv[3]  # <DeviceId>
                _log.info(u'%r connected', boxid)
            elif pv[1] == u'LOOPID':
                _log.info(u'Config Loop ID: %s', pv[3])
            elif pv[1] == u'POWER':
                _log.info(u'Config Loop Power: %s', pv[3])

    def _protocolmsg(self, pv):
        """Respond to protocol message from decoder."""
        if len(pv) == 3:
            if _RRS_PROTOCOL > pv[2]:
                _log.error(
                    u'Protocol %r unsupported (max %r), update firmware',
                    _RRS_PROTOCOL, pv[2])

    def _passingsmsg(self, pv):
        """Handle update of current passing count."""
        try:
            newfile = None
            if len(pv) > 1:
                newfile = int(pv[1])
            newidx = int(pv[0])
            if newfile != self._curfile:
                _log.debug(u'New passing file %r', newfile)
                self._curfile = newfile
                self._lastpassing = newidx
            else:
                if self._lastpassing is None or newidx < self._lastpassing:
                    # assume new connection or new file
                    _log.debug(u'Last passing %r updated to %r',
                               self._lastpassing, newidx)
                    self._lastpassing = newidx
                else:
                    # assume a broken connection and fetch missed passings
                    _log.debug(u'Missed %r passings, last passing = %r',
                               newidx - self._lastpassing, self._lastpassing)
            self._fetchpending = False
        except Exception as e:
            _log.debug(u'%s reading passing count: %s', e.__class__.__name__,
                       e)

    def _procmsg(self, msg):
        """Process a decoder response message."""
        #_log.debug(u'RECV: %r', msg)
        mv = msg.strip().split(u';')
        if mv[0].isdigit():  # Requested passing
            self._pending_command = u'PASSING'
            self._passing(mv)
        elif mv[0] == u'#P':  # Pushed passing
            self._passing(mv[1:])
        elif mv[0] == u'GETSTATUS':
            self._statusmsg(mv[1:])
        elif mv[0] == u'GETCONFIG':
            self._configmsg(mv[1:])
        elif mv[0] == u'GETPROTOCOL':
            self._protocolmsg(mv[1:])
        elif mv[0] == u'PASSINGS':
            self._passingsmsg(mv[1:])
        elif mv[0] == u'SETTIME':
            _log.info(u'Time set to: %r %r', mv[1], mv[2])
        elif mv[0] == u'STARTOPERATION':
            self._curfile = None
            self._lastpassing = None
            _log.info(u'Start session')
        elif mv[0] == u'STOPOPERATION':
            self._curfile = None
            self._lastpassing = None
            _log.info(u'Stop session')
        elif mv[0].startswith(u'ONLY '):
            self._fetchpending = False
        elif mv[0] == u'':
            _log.debug(u'End of requested passings')
            self._fetchpending = False
            self._pending_command = None
            self._dorefetch = True
        else:
            pass  # Ignore other responses

    def _procline(self):
        """Read and process whole line from decoder, return command status."""
        idx = self._rdbuf.find(b'\n')
        if idx < 0:
            inb = self._io.recv(512)
            if inb == b'':
                _log.info(u'Connection closed by peer')
                self._close()
            else:
                self._rdbuf += inb
            idx = self._rdbuf.find(b'\n')
        if idx >= 0:
            l = self._rdbuf[0:idx + 1].decode(_RRS_ENCODING)
            self._rdbuf = self._rdbuf[idx + 1:]
            self._procmsg(l)
        return self._pending_command is None

    def _refetch(self):
        """Poll decoder for new passings."""
        if not self._fetchpending:
            if self._dorefetch and self._lastpassing is not None:
                self._fetchpending = True
                self._dorefetch = False
                cmd = u'{:d}:32'.format(self._lastpassing + 1)
                self.write(cmd)

    def run(self):
        """Decoder main loop."""
        _log.debug(u'Starting')
        if sysconf.has_option(u'rrs', u'allowstored'):
            self._allowstored = sysconf.get_bool(u'rrs', u'allowstored')
            _log.debug('Allow stored passings: %r', self._allowstored)
        if sysconf.has_option(u'rrs', u'passiveloop'):
            self._passiveloop = sysconf.get_str(u'rrs', u'passiveloop', u'1')
            _log.info(u'Passive loop id set to: %r', self._passiveloop)
        self._running = True
        while self._running:
            try:
                c = None
                if self._io is not None:
                    # Read responses until response complete or timeout
                    try:
                        while not self._procline():
                            pass
                    except socket.timeout:
                        self._dorefetch = True
                    self._refetch()
                    c = self._cqueue.get_nowait()
                else:
                    c = self._cqueue.get()
                self._cqueue.task_done()
                self._proccmd(c)
            except Queue.Empty:
                pass
            except socket.error as e:
                self._close()
                _log.error(u'%s: %s', e.__class__.__name__, e)
            except Exception as e:
                _log.critical(u'%s: %s', e.__class__.__name__, e)
                self._running = False
        self.setcb()
        _log.debug(u'Exiting')
